'''
Module implementing a general Brain-Computer Interface which manages
and incoming stream of neural data and responds to it in real-time.
'''

import time

import numpy as np


class generic_BCI:
    '''
    Implements a generic Brain-Computer Interface.

    Internally manages a buffer of the signal given in `inlet` and
    continuously performs classification on appropriately transformed
    real-time neural data. At each classification, a corresponding `action`
    is performed.

    Attributes:
        classifier: a function which performs classification on the
            most recent data (transformed as needed). returns classification.
        transformer: function which takes in the most recent data (`buffer`)
            and returns the transformed input the classifer expects.
        action: a function which takes in the classification, and
            performs some action.
        calibrator: a function which is run on startup to perform
            calibration using `inlet`; returns `calibration_info`
            which is used by `classifier` and `transformer`.
        calibration_info: the result of the calibrator, if applicable.
        buffer_length(int): the length of the `buffer`; specifies the
            number of samples of the signal to keep for classification.
        brain_state: the most recent brain state classification.

    '''

    def __init__(self, classifier, transformer=None, action=print,
                 calibrator=None, buffer_length=1024):
        '''
        Initialize a generic BCI object.

        See class documentation for infromation about the class itself.

        Arguments:
            classifier: a function which performs classification on the
                most recent data (transformed as needed). returns class.
            transformer: function which takes in the most recent data (`buffer`)
                and returns the transformed input the classifer expects.
            action: a function which takes in the classification, and
                performs some action.
            calibrator: a function which is run on startup to perform
                calibration using `inlet`; returns `calibration_info`
                which is used by `classifier` and `transformer`.
            buffer_length(int): the length of the `buffer`; specifies the
                number of samples of the signal to keep for classification.
        '''

        self.brain_state = None

        self.classifier = classifier
        self.transformer = transformer
        self.action = action

        self.calibrator = calibrator
        self.calibration_info = None

        self.buffer_length = buffer_length

    def calibrate(self, inlet):
        '''
        runs the `calibrator`.

        return value of `calibrator` is stored in the instance's
        `calibration_info` which the `transformer` and `classifier`
        can use at run-time of BCI.

        Arguments:
            inlet: a pylsl `StreamInlet` of the brain signal.
        '''
        # run calibrator to get `calibration_info` which is used when
        # perfroming transformation and classifiaction
        if self.calibrator is not None:
            self.calibration_info = self.calibrator(inlet)
        else:
            print('Instance of generic BCI has no calibrator')
            self.calibration_info = None

    def _update(self, buffer):

        if buffer.shape[0] > self.buffer_length:
            # clip to buffer_length
            buffer = buffer[-self.buffer_length:]

            # transform buffer for classification
            if self.transformer is not None:
                try:
                    clf_input = self.transformer(buffer, self.calibration_info)
                except TypeError as type_err:
                    print(('Got TypeError when calling transformer.\n\n'
                           'Make sure your transformer is a function '
                           'which accepts (buffer, calibration_info) as inputs.'
                           ))
                    raise type_err
            else:
                clf_input = buffer

            try:
                # perform classification
                self.brain_state = self.classifier(clf_input,
                                                   self.calibration_info)
            except TypeError as type_err:
                print(('Got TypeError when calling classifier. \n\n'
                       'Make sure your classifier is a function which '
                       'accepts (clf_input, calibration_info) as inputs.'))
                raise type_err

            try:
                # run action based on classification
                self.action(self.brain_state)
            except TypeError as type_err:
                print(('Got TypeError when calling action. \n\n'
                       'Make sure your action is a function which '
                       'accepts (brain_state) as an input.'))
                raise type_err

    def run(self, inlet):
        '''
        Runs the defined Brain-Computer Interface.

        Internally manages a buffer of the signal given in `inlet` and
        continuously performs classification on appropriately transformed
        real-time neural data. At each classification, a corresponding `action`
        is performed.


        Arguments:
            inlet: a pylsl `StreamInlet` of the brain signal.
        '''

        inlet.open_stream()

        # get number of available channels in inlet
        n_channels = inlet.channel_count
        buffer = np.empty((0, n_channels))  # initialize buffer

        # TODO: implement ending condition?
        running = True  # currently constantly running.

        while running:
            chunk, _ = inlet.pull_chunk(max_samples=self.buffer_length)
            if np.size(chunk) != 0:  # Check if new data available
                buffer = np.append(buffer, np.array(chunk), axis=0)

                self._update(buffer)

    def test_update_rate(self, inlet, test_length=10, perform_action=True):
        '''
        Returns the rate at which the BCI is able to make a classification
        and perform its action.

        Arguments:
            inlet: a pylsl `StreamInlet` of the brain signal.
            test_length(float): how long to run the test for in seconds.
            perform_action(bool): whether to perform the action or skip it.
        '''

        inlet.open_stream()

        # get number of available channels in inlet
        n_channels = inlet.channel_count
        buffer = np.empty((0, n_channels))  # initialize buffer

        n_updates = 0
        start_time = time.time()

        while time.time() - start_time < test_length:
            chunk, _ = inlet.pull_chunk(max_samples=self.buffer_length)
            if np.size(chunk) != 0:  # Check if new data available
                buffer = np.append(buffer, np.array(chunk), axis=0)

                self._update(buffer)

                n_updates += 1

        update_rate = n_updates / test_length

        return update_rate


class fsm_BCI(generic_BCI):
    '''
    Implements a Finite-State-Machine-inspired Brain-Computer Interface.

    Classification of brain-state is not only dependent on the transformed
    real-time brain signal, but also the previous brain state.

    Internally manages a buffer of the signal given in `inlet` and
    continuously performs classification on appropriately transformed
    real-time neural data. At each classification, a corresponding `action`
    is performed.

    Attributes:
        classifier: a function which performs classification on the
            most recent data (transformed as needed). returns classification.
        transformer: function which takes in the most recent data (`buffer`)
            and returns the transformed input the classifer expects.
        action: a function which takes in the classification, and
            performs some action.
        calibrator: a function which is run on startup to perform
            calibration using `inlet`; returns `calibration_info`
            which is used by `classifier` and `transformer`.
        calibration_info: the result of the calibrator, if applicable.
        buffer_length(int): the length of the `buffer`; specifies the
            number of samples of the signal to keep for classification.
        brain_state: the most recent brain state classification.

    '''

    def _update(self, buffer):
        if buffer.shape[0] > self.buffer_length:
            # clip to buffer_length
            buffer = buffer[-self.buffer_length:]

            # transform buffer for classification
            if self.transformer is not None:
                try:
                    clf_input = self.transformer(buffer, self.calibration_info)
                except TypeError as type_err:
                    print(('Got TypeError when calling transformer.\n\n'
                           'Make sure your transformer is a function '
                           'which accepts (buffer, calibration_info) as inputs.'
                           ))
                    raise type_err
            else:
                clf_input = buffer

            try:
                # perform classification
                self.brain_state = self.classifier(clf_input, self.brain_state,
                                                   self.calibration_info)
            except TypeError as type_err:
                print(('Got TypeError when calling classifier. \n\n'
                       'Make sure your classifier is a function which accepts '
                       '(clf_input, brain_state, calibration_info) as inputs.'))
                raise type_err

            try:
                # run action based on classification
                self.action(self.brain_state)
            except TypeError as type_err:
                print(('Got TypeError when calling action. \n\n'
                       'Make sure your action is a function which '
                       'accepts (brain_state) as an input.'))
                raise type_err


class retentive_BCI(generic_BCI):
    '''
    Implements a Brain-Computer Interface with memory of past brain states.

    Classification of brain-state is not only dependent on the transformed
    real-time brain signal, but also the finite list of previous brain states.

    Internally manages a buffer of the signal given in `inlet` and
    continuously performs classification on appropriately transformed
    real-time neural data. At each classification, a corresponding `action`
    is performed.

    Attributes:
        classifier: a function which performs classification on the
            most recent data (transformed as needed). returns classification.
        transformer: function which takes in the most recent data (`buffer`)
            and returns the transformed input the classifer expects.
        action: a function which takes in the classification, and
            performs some action.
        calibrator: a function which is run on startup to perform
            calibration using `inlet`; returns `calibration_info`
            which is used by `classifier` and `transformer`.
        calibration_info: the result of the calibrator, if applicable.
        buffer_length(int): the length of the `buffer`; specifies the
            number of samples of the signal to keep for classification.
        brain_state: the most recent brain state classification.
        memory_length(int): number of brain states into the past to remember.
        past_states: a list of the past classifications of brain states.
            used in next classification. length is memory_length.
    '''

    def __init__(self, classifier, transformer=None, action=print,
                 calibrator=None, buffer_length=1024, memory_length=10):
        '''
        Initialize a generic BCI object.

        See class documentation for infromation about the class itself.

        Arguments:
            classifier: a function which performs classification on the
                most recent data (transformed as needed). returns class.
            transformer: function which takes in the most recent data (`buffer`)
                and returns the transformed input the classifer expects.
            action: a function which takes in the classification, and
                performs some action.
            calibrator: a function which is run on startup to perform
                calibration using `inlet`; returns `calibration_info`
                which is used by `classifier` and `transformer`.
            buffer_length(int): the length of the `buffer`; specifies the
                number of samples of the signal to keep for classification.
            memory_length(int): number of brain states to remember into past.
        '''

        super().__init__(classifier,
                         transformer, action, calibrator, buffer_length)

        self.memory_length = memory_length
        self.past_states = []

    def _update(self, buffer):
        if buffer.shape[0] > self.buffer_length:
            # clip to buffer_length
            buffer = buffer[-self.buffer_length:]

            # transform buffer for classification
            if self.transformer is not None:
                try:
                    clf_input = self.transformer(buffer, self.calibration_info)
                except TypeError as type_err:
                    print(('Got TypeError when calling transformer.\n\n'
                           'Make sure your transformer is a function '
                           'which accepts (buffer, calibration_info) as inputs.'
                           ))
                    raise type_err
            else:
                clf_input = buffer

            try:
                # perform classification
                self.brain_state = self.classifier(clf_input, self.past_states,
                                                   self.calibration_info)
                self.past_states.append(self.brain_state)
                if len(self.past_states) > self.memory_length:
                    self.past_states = self.past_states[-self.memory_length:]
            except TypeError as type_err:
                print(('Got TypeError when calling classifier. \n\n'
                       'Make sure your classifier is a function which accepts '
                       '(clf_input, past_states, calibration_info) as inputs.'))
                raise type_err

            try:
                # run action based on classification
                self.action(self.brain_state)
            except TypeError as type_err:
                print(('Got TypeError when calling action. \n\n'
                       'Make sure your action is a function which '
                       'accepts (brain_state) as an input.'))
                raise type_err



'''
Module containing functions to study and analyze neural signals,
especially to provide insights for building machine learning models
to perform classification relevant to Brain-Computer Interface applications.
'''

import matplotlib.pyplot as plt
import numpy as np
from sklearn.decomposition import PCA, FastICA


def plot_signal(signal, sampling_rate, signal_type=None, ch_names=None,
                event_timestamps=None, **plt_kwargs):
    '''
    Plots signal.

    Arguments:
        signal: signal as array of shape [n_samples, n_channels].
        sr(float): sampling rate in samples per second.
        signal_type: (optional) gives a title for the y-axis.
        ch_names: (optional) array of names for each channel (used for legend).
        event_timestamps: (optional) 1-D array of times at which an
            event/stimulus occured.
        **plt_kwargs: matplotlib keyword args
    '''

    time = np.linspace(0, len(signal)/sampling_rate, len(signal))

    plt.plot(time, signal, **plt_kwargs)

    if ch_names is not None:
        plt.legend(ch_names)

    if event_timestamps is not None:
        for timestamp in event_timestamps:
            plt.axvline(timestamp, ls='--', lw=1, c='violet')

    plt.xlabel("time [s]")
    plt.ylabel(signal_type)


def plot_grid(signals, num_signals=None, sampling_rate=1, cols=4,
              fig_size=(10, 6), sharey=True, sharex=True, random=True,
              fig_axes=None, show=True):
    '''
    Plot an (optionally random) set of signals [epochs] in a grid
    from a larger array of given signals.

    Arguments:
        signals: array of signals to plot from (num_signals, num_samples).
        num_signals(int): the number of siganls to plot.
        sr(float): sampling rate of signals.
        cols(int): the number of columns in the grid.
        fig_size: tuple (x,y) of figure size in inches.
        sharey(bool): whether to share scale on y-axis (see matplotlib).
        sharex(bool): whether to share scale on x-axis (see matplotlib).
        random(bool): whether to choose signals randomly or just use
            the first num_signals.
        fig_axes: optionally, an existing tuple of (fig,axes) to plot on
            (see matplotlib) rather creating new one.
        show(bool): whether to show plot inline.

    Returns:
        fig, axes: matplotlib figure and axes with sample of signals
            plotted in a grid
    '''

    # if num_signals to be plotted is not given, plot all signals
    if num_signals is None:
        num_signals = signals.shape[0]

    rows = int(np.ceil(num_signals / cols))

    if fig_axes is None:
        fig, axes = plt.subplots(
            rows, cols, sharey=sharey, sharex=sharex, figsize=fig_size)
    else:
        fig, axes = fig_axes

    if random:
        # choose a random set of signals to plot
        sampled_signals = signals[np.random.choice(
            range(len(signals)), num_signals, replace=False)]
    else:
        sampled_signals = signals[:num_signals]  # choose the first num

    for row in range(rows):
        for col in range(cols):
            if (row*cols + col) >= num_signals:
                break
            sampled_signal = sampled_signals[row*cols+col]

            t_max = len(sampled_signal) / sampling_rate
            time = np.linspace(0, t_max, len(sampled_signal))
            axes[row][col].plot(time, sampled_signal)

    if show:
        plt.show()
    return fig, axes


def stim_triggered_average(signal, sampling_rate, timestamps, duration_before,
                           duration_after, plot=False):
    '''
    Inspired by the computational neuroscience concept of the
    spike-triggered average, this function computes the average
    signal characteristic around known events.

    Arguments:
        signal: signal in the form of an array of shape [samples, channels].
        sr(float): sampling rate of the signal.
        timestamps: array of floats containing the timestamps for each event.
        duration_before: the duration to be considered before each event.
        duration_after: the duration to be considered after each event.

    Returns:
        stim_triggered_average: average signal characteristic around event.
        relative_time: relative time of each sample in stim_triggered_average
            with respect to event.

    '''

    stim_indices = (timestamps*sampling_rate).astype(int)

    ind_before = int(sampling_rate * duration_before)
    ind_after = int(sampling_rate * duration_after)

    sta = np.mean(
        [signal[i-ind_before:i+ind_after] for i in stim_indices], axis=0)
    sta = np.mean(
        [signal[i-ind_before:i+ind_after] for i in stim_indices], axis=0)
    relative_time = np.linspace(-duration_before,
                                duration_after, len(sta))

    if plot:
        plt.plot(relative_time, sta)
        plt.axvline(0, c='violet', ls='--', lw=2)

    return sta, relative_time


def plot_PCA(epochs, sampling_rate=1, n_components=None, return_PCA=False,
             PCA_kwargs=None, plot_grid_kwargs=None):
    '''
    performs independent component analysis and plots independent components
    of epochs of a signal.

    Arguments:
        epochs: array of epochs (n_epochs, n_samples).
        sr(float): sampling rate.
        num_components(int): number of components to use.
            If none is passed, all are used.
        return_PCA(bool): whether to return the independent components.
        PCA_kwargs(dict): dictionary containing kwargs for PCA function
            (see scikit-learn).
        plot_grid_kwargs(dict): dictionary contianing kwargs for
            plot_grid function.

    '''
    if not PCA_kwargs:
        # TODO: change to be done like in classification_tools.transform
        PCA_kwargs = {}
    pca = PCA(n_components=n_components, **PCA_kwargs)
    principle_components = pca.fit_transform(epochs.T).T

    if not plot_grid_kwargs:
        # TODO: change to be done like in classification_tools.transform
        plot_grid_kwargs = {}
    plot_grid(principle_components, sr=sampling_rate,
              random=False, **plot_grid_kwargs)

    if return_PCA:
        return principle_components


def plot_ICA(epochs, sampling_rate=1, n_components=None, return_ICA=False,
             FastICA_kwargs=None, plot_grid_kwargs=None):
    '''
    performs independent component analysis and plots independent components of
    epochs of a signal.

    Arguments:
        epochs: array of epochs (n_epochs, n_samples).
        sr(float): sampling rate.
        num_components(int): number of components to use.
            If none is passed, all are used.
        return_ICA(bool): whether to return the independent components.
        FastICA_kwargs(dict): dictionary containing kwargs for FastICA
            function (see scikit-learn).
        plot_grid_kwargs(dict): dictionary contianing kwargs for
            plot_grid function.

    '''

    if not FastICA_kwargs:
        # TODO: change to be done like in classification_tools.transform
        FastICA_kwargs = {}
    ICA = FastICA(n_components=n_components, **FastICA_kwargs)
    independent_components = ICA.fit_transform(epochs.T).T

    if not plot_grid_kwargs:
        # TODO: change to be done like in classification_tools.transform
        plot_grid_kwargs = {}
    plot_grid(independent_components, sr=sampling_rate,
              random=False, **plot_grid_kwargs)

    if return_ICA:
        return independent_components



'''
Module for managing the models which come pre-packaged with neurol.
Includes functionality for importing and using the models.
'''

import os

from keras.models import load_model

# Note: currently all models are keras .h5 models.
# Think about organization of this module as new models are added



# get path to module to compute path to models
here = os.path.dirname(os.path.realpath(__file__))

def get_model(model_name):
    '''
    gets the specified trained model.

    Arguments:
        model_name(str): name of model.
            See documentation for list of available models.

    Returns:
        model: trained model.of
    '''

    path_to_model = here + '/' + model_name + '.h5'
    model = load_model(path_to_model)
    return model


def get_predictor(model_name):
    '''
    gets the predictor for the specified model.

    Arguments:
        model_name(str): name of model.
            See documentation for list of available models.

    Returns:
        predictor: predictor of trained model.
    '''

    model = get_model(model_name)
    predictor = model.predict
    return predictor



'''
Module including utility functions for creating `classifier`'s,
`transfromer`'s, and `calibrator`'s for use in the `BCI` module.
'''

import time

import biosppy.signals as bsig
import numpy as np

from .models.classification_tools import get_channels
from .models.preprocessing import epoch, epoch_band_features

# is this right? is it 220 Hz (see documentation)?
DEVICE_SAMPLING_RATE = {'muse': 256,
                        }


def ensemble_transform(buffer, epoch_len, channels, device='muse',
                       transformers=None, filter_=False, filter_kwargs=None):
    '''
    Ensemble transform function. Takes in buffer as input. Extracts the
    appropriate channels and samples, performs filtering, and transforms.

    Arguments:
        buffer: most recent stream data. shape: [n_samples, n_channels]
        epoch_len: length of epoch expected by predictor (number of samples).
        channels: list of channels expected by predictor. See get_channels.
        device: string of device name. used to get channels and sampling_rate.
        filter_: boolean of whether to perform filtering
        filter_kwargs: dictionary of kwargs to be passed to filtering function.
            See biosppy.signals.tools.filter_signal. by default,
            an order 8 bandpass butter filter is performed between 2Hz and 40Hz.
    '''

    # get the latest epoch_len samples of the buffer
    transformed_signal = np.array(buffer[-epoch_len:, :])

    # get the selected channels
    transformed_signal = get_channels(transformed_signal, channels, device)

    # filter_signal
    if filter_:
        # create dictionary of kwargs for filter_signal
        filt_kwargs = {'sampling_rate': DEVICE_SAMPLING_RATE[device],
                       'ftype': 'butter',
                       'band': 'bandpass',
                       'frequency': (2, 40),
                       'order': 8}

        if filter_kwargs is not None:
            filt_kwargs.update(filter_kwargs)

        transformed_signal, _, _ = bsig.tools.filter_signal(
            signal=transformed_signal.T, **filt_kwargs)
        transformed_signal = transformed_signal.T

    return transformed_signal


def band_power_calibrator(inlet, channels, device, bands, percentile=50,
                          recording_length=10, epoch_len=1,
                          inter_window_interval=0.2):
    '''
    Calibrator for `generic_BCI.BCI` which computes a given `percentile` for
    the power of each frequency band across epochs channel-wise. Useful for
    calibrating a concentration-based BCI.

    Arguments:
        inlet: a pylsl `StreamInlet` of the brain signal.
        channels: array of strings with the names of channels to use.
        device(str): device name for use by `classification_tools`
        bands: the frequency bands to get power features for.
            'all': all of ['theta', 'alpha_low', 'alpha_high', 'beta', 'gamma']
            otherwise an array of strings of the desired bands.
        percentile: the percentile of power distribution across epochs to
            return for each band.
        recording_length(float): length of recording to use for calibration
            in seconds.
        epoch_len(float): the length of each epoch in seconds.
        inter_window_interval(float): interval between each window/epoch
            in seconds (measured start to start)

    Returns:
        clb_info: array of shape [n_bands, n_channels] of the `percentile`
        of the power of each band
    '''
    sampling_rate = DEVICE_SAMPLING_RATE[device]  # get sampling_rate
    # calculate window size in # of samples
    ws = int(epoch_len * sampling_rate)
    # calculate inter_window_interval in # of samples
    iwi = int(inter_window_interval * sampling_rate)

    input('Press Enter to begin calibration...')

    print(f'Recording for {recording_length} seconds...')

    # sleep for recording_length while inlet accumulates chunk
    # necessary so that no data is used before the indicated start of recording
    inlet.open_stream()
    time.sleep(recording_length)
    recording, _ = inlet.pull_chunk(
        max_samples=sampling_rate*recording_length)  # get accumulated data
    # get appropriate channels
    recording = get_channels(np.array(recording), channels, device=device)

    # epoch the recording to compute percentile across distribution
    epochs = epoch(recording, ws, iwi)

    # compute band power for each epoch
    band_power = np.array([epoch_band_features(epoch, sampling_rate,
                                               bands=bands, return_dict=False)
                           for epoch in epochs])

    # calculate given percentile of band power
    clb_info = np.squeeze(np.percentile(band_power, percentile, axis=0))

    print(f'\nComputed the following power percentiles: \n{clb_info}')
    input('\nCalibration complete. Press Enter to start BCI...')

    return clb_info


def band_power_transformer(buffer, clb_info, channels, device, bands,
                           epoch_len=1):
    '''
    Transformer for `generic_BCI.BCI` which chooses channels, epochs, and
    gets power features on some choice of bands.

    Arguments:
        buffer: most recent stream data. shape: [n_samples, n_channels]
        clb_info: not used. included for compatibility with generic_BCI.BCI
        channels: list of strings of the channels to use.
        device:(str): device name for use by `classification_tools`.
        bands: the frequency bands to get power features for.
            'all': all of ['theta', 'alpha_low', 'alpha_high', 'beta', 'gamma']
            otherwise a list of strings of the desired bands.
        epoch_len(float): the duration of data to classify on in seconds.

    Returns:
        transformed_signal: array of shape [n_bands, n_channels] of the
        channel-wise power of each band over the epoch.
    '''
    sr = DEVICE_SAMPLING_RATE[device]  # get device sampling rate

    # get the latest epoch_len samples from the buffer
    transformed_signal = np.array(buffer[-int(epoch_len*sr):, :])

    # get the selected channels
    transformed_signal = get_channels(transformed_signal, channels, device)

    # compute band_features on signal
    transformed_signal = np.squeeze(epoch_band_features(
        transformed_signal, sr, bands=bands, return_dict=False))

    return transformed_signal



'''
Module containing functions for quickly connecting to BCI-related streaming
devices.
'''

import ble2lsl
from ble2lsl.devices import muse2016

from pylsl import StreamInlet, resolve_byprop


def connect_muse():
    '''
    connects to any available muse headset.
    returns ble2lsl.ble2lsl.Streamer object.
    '''
    return ble2lsl.Streamer(muse2016)


def get_EEG_inlets():
    '''resolves all EEG streams and returns their inlets in an array.'''

    stream_infos = resolve_byprop("type", "EEG")

    inlets = []

    for stream_info in stream_infos:
        inlets.append(StreamInlet(stream_info, ))

    return inlets



'''
Module containing functions for the preparation of neural data for use with
with BCI-related models.

'''

import numpy as np
from scipy import stats
import biosppy.signals as bsig


def epoch(signal, window_size, inter_window_interval):
    '''
    Creates overlapping windows/epochs of EEG data from a single recording.

    Arguments:
        signal: array of timeseries EEG data of shape [n_samples, n_channels]
        window_size(int): desired size of each window in number of samples
        inter_window_interval(int): interval between each window in number of
            samples (measured start to start)

    Returns:
        numpy array object with the epochs along its first dimension
    '''

    # Calculate the number of valid windows of the specified size in signal
    num_windows = 1 + (len(signal) - window_size) // inter_window_interval

    epochs = []
    for i in range(num_windows):
        epochs.append(signal[i*inter_window_interval:i *
                             inter_window_interval + window_size])

    return np.array(epochs)


def labels_from_timestamps(timestamps, sampling_rate, length):
    '''takes an array containing timestamps (as floats) and
    returns a labels array of size 'length' where each index
    corresponding to a timestamp via the 'samplingRate'.

    Arguments:
        timestamps: an array of floats containing the timestamps for each
        event (units matching sampling_rate).
        sampling_rate(float): the sampling rate of the EEG data.
        length(int): the number of samples of the corresponing EEG recording.

    Returns:
        an integer array of size 'length' with a '1' at each time index where a
        corresponding timestamp exists, and a '0' otherwise.
    '''
    # NOTE:  currently assumes binary labels. Generalize to arbitrary labels.

    # create labels array template
    labels = np.zeros(length)

    # calculate corresponding index for each timestamp
    labelIndices = np.array(timestamps * sampling_rate)
    labelIndices = labelIndices.astype(int)

    # flag label at each index where a corresponding timestamp exists
    labels[labelIndices] = 1
    labels = labels.astype(int)

    return np.array(labels)


def label_epochs(labels, window_size, inter_window_interval, label_method):
    '''
    create labels for individual eoicgs of EEG data based on the
    label_method.

    Arguments:
        labels: an integer array indicating a class for each sample measurement
        window_size(int): size of each window in number of samples
            (matching window_size in epoched data)
        inter_window_interval(int): interval between each window in number of
            samples (matching inter_window_interval in epoched data)
        label_method(str/func): method of consolidating labels contained in
            epoch into a single label.
                'containment': whether a positive label occurs in the epoch,
                'count': the count of positive labels in the epoch,
                'mode': the most common label in the epoch
                func: func_name(epoched_labels) outputs label of epoched_labels

    Returns:
        a numpy array with a label correponding to each epoch
    '''

    # epoch the labels themselves so each epoch contains a label at each sample
    epochs = epoch(labels, window_size, inter_window_interval)

    # if a positive label [1] occurs in the epoch, give epoch positive label
    if label_method == 'containment':
        epoch_labels = [int(1 in epoch) for epoch in epochs]

    elif label_method == 'count':
        # counts the number of occurences of the positive label [1]]
        epoch_labels = [epoch.count(1) for epoch in epochs]

    elif label_method == 'mode':
        # choose the most common label occurence in the epoch
        # default to the smallest if multiple exist
        epoch_labels = [stats.mode(epoch)[0][0] for epoch in epochs]

    elif callable(label_method):
        epoch_labels = [label_method(epoch) for epoch in epochs]

    return np.array(epoch_labels)


def label_epochs_from_timestamps(timestamps, sampling_rate, length,
                                 window_size, inter_window_interval,
                                 label_method='containment'):
    '''
    Directly creates labels for individual windows of EEG data from
    timestamps of events.

    Arguments:
        timestamps: an array of floats containing the timestamps
            for each event (units matching sampling_rate).
        sampling_rate(float): sampling rate of the recording.
        length(int): the number of samples of the corresponing EEG recording.
        window_size(int): size of each window in number of samples
            (matches window_size in epoched data)
        inter_window_interval(int): interval between each window in number of
            samples (matches inter_window_interval in epoched data)
        label_method(str/func): method of consolidating labels contained in
            epoch into a single label.
                'containment': whether a positive label occurs in the epoch,
                'count': the count of positive labels in the epoch,
                'mode': the most common label in the epoch
                func: func_name(epoched_labels) outputs label of epoched_labels

    Returns:
        an array with a label correponding to each window
    '''
    labels = labels_from_timestamps(timestamps, sampling_rate, length)

    epoch_labels = label_epochs(
        labels, window_size, inter_window_interval, label_method)

    return epoch_labels


def epoch_and_label(data, sampling_rate, timestamps, window_size,
                    inter_window_interval, label_method='containment'):
    '''
    Epochs a signal (single EEG recording) and labels each epoch using
    timestamps of events and a chosen labelling method.

    Arguments:
        data: array of timeseries EEG data of shape [n_samples, n_channels]
        timestamps: an array of floats containing the timestamps for each event
            in units of time.
        sampling_rate(float): the sampling rate of the EEG data.
        window_size(float): desired size of each window in units of time.
        inter_window_interval(float): interval between each window
            in units of time (measured start to start)
        label_method(str/func): method of consolidating labels contained in
            epoch into a single label.
                'containment': whether a positive label occurs in the epoch,
                'count': the count of positive labels in the epoch,
                'mode': the most common label in the epoch
                func: func_name(epoched_labels) outputs label of epoched_labels

    Returns:
        epochs: array of epochs with shape [n_epochs, n_channels]
        labels: array of labels corresponding to each epoch of shape [n_epochs,]
    '''

    ws = int(window_size*sampling_rate)
    iwi = int(inter_window_interval*sampling_rate)
    epochs = epoch(data, ws, iwi)
    labels = label_epochs_from_timestamps(timestamps, sampling_rate, len(data),
                                          ws, iwi, label_method=label_method)

    return epochs, labels


def compute_signal_std(signal, corrupt_intervals=None, sampling_rate=1):
    '''
    Computes and returns the standard deviation of a signal channel-wise
    while avoiding corrupt intervals

    Arguments:
        signal: signal of shape [n_samples, n_channels]
        corrupt_intervals: an array of 2-tuples indicating the start and
            end time of the corrupt interval (units of time)
        sampling_rate: the sampling rate in units of samples/unit of time


    Returns:
        standard deviation of signal channel-wise of shape [1, n_channels]
    '''

    # convert signal into numpy array for use of its indexing features
    signal = np.array(signal)

    if corrupt_intervals is not None:
        # convert corrupt_indices from units of time to # of samples
        corrupt_intervals = [(int(cor_start * sampling_rate),
                              int(cor_end * sampling_rate))
                             for cor_start, cor_end in corrupt_intervals]

        # find all indices to keep
        good_indices = []

        # for each index in the signal,
        # check if it is contained in any of the corrupt intervals
        for ind in range(len(signal)):
            good_indices.append(not np.any(
                [ind in range(cor_start, cor_end)
                 for cor_start, cor_end in corrupt_intervals]))

        signal = signal[good_indices]

    # compute and return std on non_corrupt parts of signal
    return np.std(signal, axis=0, dtype=np.float32)


def split_corrupt_signal(signal, corrupt_intervals, sampling_rate=1):
    '''
    Splits a signal with corrupt intervals and returns array of signals
    with the corrupt intervals filtered out. This is useful for treating
    each non_corrupt segment as a seperate signal to ensure continuity
    within a single signal.

    Arguments:
        signal: signal of shape [n_samples, n_channels]
        corrupt_intervals: an array of 2-tuples indicating the start and
            end time of the corrupt interval (units of time)
        sampling_rate: the sampling rate in units of samples/unit of time


    Returns:
        array of non_corrupt signals of shape [n_signal, n_samples, n_channels]
    '''

    # convert corrupt_indices from units of time and flatten into single array
    corrupt_intervals = np.array(corrupt_intervals * sampling_rate).flatten()

    # convert signal into numpy array for use with library
    signal = np.array(signal)

    # split signal on each cor_start/cor_end and discard the regions in between
    # cor_start and cor_end
    signals = np.split(signal, corrupt_intervals)[::2]

    return signals


def epoch_band_features(epoch_, sampling_rate, bands='all', return_dict=True):
    '''
    Computes power features of EEG frequency bands over the epoch channel-wise.

    Arguments:
        epoch_: a single epoch of shape [n_samples, n_channels]
        sampling_rate: the sampling rate of the signal in units of samples/sec
        bands: the requested frequency bands to get power features for.
            'all': all of ['theta', 'alpha_low', 'alpha_high', 'beta', 'gamma']
            otherwise an array of strings of the desired bands.
        return_dict(bool): returns `band_features` in the form of a dictionary
            if True, else returns as numpy array in order of `bands`

    Returns:
        a dictionary of arrays of shape [1, n_channels] containing the
        power features over each frequency band per channel.
    '''

    if bands == 'all':
        bands = ['theta', 'alpha_low', 'alpha_high', 'beta', 'gamma']

    # computes length of epoch to compute band power features over entire epoch
    w_size = np.shape(epoch_)[0] / sampling_rate

    # compute power for each band
    _, theta, alpha_low, alpha_high, beta, gamma = bsig.eeg.get_power_features(
        epoch_, sampling_rate=sampling_rate, size=w_size, overlap=0)

    # initialize return dict
    band_features = {'theta': theta,
                     'alpha_low': alpha_low,
                     'alpha_high': alpha_high,
                     'beta': beta,
                     'gamma': gamma}

    # return only requested bands
    # (intersection of bands available and bands requested)
    if return_dict:  # return in form of dict
        return_band_features = {
            band: band_features[band] for band in bands & band_features.keys()}
    else:
        return_band_features = np.array(
            [band_features[band] for band in bands & band_features.keys()])

    return return_band_features



'''
Module for plotting stream of neural data.
Includes time domain, fourrier transform, and spectrogram live plots.
'''

import numpy as np
from scipy import signal
import pyqtgraph as pg


#TODO:
# add filtering

def plot(inlet, buffer_length=2048, channels=None, w_size=(1920, 1080)):
    """
    plots data stream from inlet. one row per channel.

    Args:
        inlet (pylsl.pylsl.StreamInlet): the pylsl inlet streaming neural data.
        buffer_length (int, optional): length of buffer to plot (# of samples).
            Defaults to 2048.
        channels: channels to plot. list/tuple of channel indices,
            or dict with indices as keys and names as values.
            Defaults to None (plots all channels w/o names).
        w_size (tuple, optional): initial size of window in pixels.
            Defaults to (1920, 1080).
    """


    # get sampling rate and number of available channels from inlet
    info = inlet.info()
    n_channels = info.channel_count()
    s_rate = info.nominal_srate()


    buffer = np.zeros((buffer_length, n_channels))  # initialize buffer
    ts = np.linspace(-int(buffer_length/s_rate), 0, num=buffer_length)

    # initialize pyqt graph app, window
    win = pg.GraphicsLayoutWidget(show=True)
    win.setWindowTitle('Live Plot')
    app = pg.QtGui.QApplication
    win.resize(w_size[0], w_size[1])

    # initialize plots/curves
    plots = []
    plot_curves = []

    if channels is None:
        channels = list(range(n_channels))

    if isinstance(channels, (list, tuple)):
        for ch_ind in channels:
            plt = win.addPlot(title=f'channel {ch_ind}',
                              labels={'left': 'mV'}
                              )
            plt.setMouseEnabled(x=False, y=False)
            plt.showGrid(x=True)

            plt_curve = plt.plot(y=buffer[:, ch_ind], x=ts)

            plots.append(plt)
            plot_curves.append(plt_curve)

            win.nextRow()

        ch_inds = list(channels)

    elif isinstance(channels, dict):
        for ch_ind, ch_name in channels.items():
            plt = win.addPlot(title=f'{ch_name} channel',
                              labels={'left': 'mV'}
                              )
            plt.setMouseEnabled(x=False, y=False)
            plt.showGrid(x=True)

            plt_curve = plt.plot(y=buffer[:, ch_ind], x=ts)

            plots.append(plt)
            plot_curves.append(plt_curve)

            win.nextRow()

        ch_inds = list(channels.keys())

    else:
        raise ValueError('`channels` argument should be list, tuple, or dict')


    # label bottom plot's x-axis
    plots[-1].setLabel('bottom', 'time', units='s')


    # process initialization events
    app.processEvents()

    # open stream
    inlet.open_stream()

    # currently always running, TODO: implement ending condition on app close
    running = True

    while running:
        chunk, _ = inlet.pull_chunk(max_samples=buffer_length)
        if np.size(chunk) != 0:  # Check if new data available
            buffer = np.append(buffer, np.array(chunk), axis=0)

            # clip to buffer_length
            buffer = buffer[-buffer_length:]

            for ch_ind, plot_curve in zip(ch_inds, plot_curves):
                plot_curve.setData(y=buffer[:, ch_ind], x=ts)

            app.processEvents()

        if not win.isVisible():
            running = False
            app.quit()

def plot_fft(inlet, buffer_length=2048, channels=None, w_size=(1920, 1080)):
    """
    plots fourrier transform of data stream from inlet. one row per channel.

    Args:
        inlet (pylsl.pylsl.StreamInlet): the pylsl inlet streaming neural data.
        buffer_length (int, optional): length of buffer to plot (# of samples).
            Defaults to 2048.
        channels: channels to plot. list/tuple of channel indices,
            or dict with indices as keys and names as values.
            Defaults to None (plots all channels w/o names).
        w_size (tuple, optional): initial size of window in pixels.
            Defaults to (1920, 1080).
    """


    # get sampling rate and number of available channels from inlet
    info = inlet.info()
    n_channels = info.channel_count()
    s_rate = info.nominal_srate()


    buffer = np.zeros((buffer_length, n_channels))  # initialize buffer
    fs = np.fft.fftshift(np.fft.fftfreq(buffer_length, 1/s_rate))

    # initialize pyqt graph app, window
    win = pg.GraphicsLayoutWidget(show=True)
    win.setWindowTitle('Fourrier Transform Live Plot')
    app = pg.QtGui.QApplication
    win.resize(w_size[0], w_size[1])

    # initialize plots/curves
    plots = []
    plot_curves = []

    if channels is None:
        channels = list(range(n_channels))

    if isinstance(channels, (list, tuple)):
        for ch_ind in channels:
            plt = win.addPlot(title=f'channel {ch_ind}',
                              labels={'left': 'mV'}
                              )
            plt.setMouseEnabled(x=False, y=False)
            plt.showGrid(x=True)

            plt_curve = plt.plot(x=fs)

            plots.append(plt)
            plot_curves.append(plt_curve)

            win.nextRow()

        ch_inds = list(channels)

    elif isinstance(channels, dict):
        for ch_ind, ch_name in channels.items():
            plt = win.addPlot(title=f'{ch_name} channel')
            plt.setLabel('left', 'voltage-secs', units='Vs') # finalize above
            plt.setMouseEnabled(x=False, y=False)
            plt.showGrid(x=True)

            plt_curve = plt.plot(x=fs)

            plots.append(plt)
            plot_curves.append(plt_curve)

            win.nextRow()

        ch_inds = list(channels.keys())

    else:
        raise ValueError('`channels` argument should be list, tuple, or dict')


    # label bottom plot's x-axis
    plots[-1].setLabel('bottom', 'frequency', units='Hz')


    # process initialization events
    app.processEvents()

    # open stream
    inlet.open_stream()

    running = True
    while running:
        chunk, _ = inlet.pull_chunk(max_samples=buffer_length)
        if np.size(chunk) != 0:  # Check if new data available
            buffer = np.append(buffer, np.array(chunk)/1000, axis=0)

            # clip to buffer_length
            buffer = buffer[-buffer_length:]

            # compute fft
            fft = np.fft.fftshift(np.fft.fft(buffer, axis=0))
            fft_mag = np.abs(fft)

            for ch_ind, plot_curve in zip(ch_inds, plot_curves):
                plot_curve.setData(y=fft_mag[:, ch_ind], x=fs)

            app.processEvents()

        if not win.isVisible():
            running = False
            app.quit()


# NOTE: this can be optimized, currentlythere are redundant computations
# NOTE: works, but spectrogram always seems to produce extreme outliers
# that mess up the graph...

def plot_spectrogram(inlet, buffer_length=2048, channels=None,
                     w_size=(1920, 1080)):
    """
    plots spectrogram of data stream from inlet. one row per channel.

    Args:
        inlet (pylsl.pylsl.StreamInlet): the pylsl inlet streaming neural data.
        buffer_length (int, optional): length of buffer to plot (# of samples).
            Defaults to 2048.
        channels: channels to plot. list/tuple of channel indices,
            or dict with indices as keys and names as values.
            Defaults to None (plots all channels w/o names).
        w_size (tuple, optional): initial size of window in pixels.
            Defaults to (1920, 1080).
    """


    # get sampling rate and number of available channels from inlet
    info = inlet.info()
    n_channels = info.channel_count()
    s_rate = info.nominal_srate()


    buffer = np.zeros((buffer_length, n_channels))  # initialize buffer
    freqs, ts, Sxx = signal.spectrogram(buffer[:, 0], s_rate, mode='magnitude')
    ts = np.max(ts) - ts

    # initialize pyqt graph app, window
    win = pg.GraphicsLayoutWidget(show=True)
    win.setWindowTitle('Live Plot')
    app = pg.QtGui.QApplication
    win.resize(w_size[0], w_size[1])

    # interpret image data as row-major to match scipy output
    pg.setConfigOptions(imageAxisOrder='row-major')


    # initialize plots/curves
    plots = []
    plot_imgs = []
    #plot_hists = []

    if channels is None:
        channels = list(range(n_channels))

    if isinstance(channels, (list, tuple)):
        for ch_ind in channels:
            plt = win.addPlot(title=f'channel {ch_ind}')
            plt.setMouseEnabled(x=False, y=False)

            img = pg.ImageItem()
            plt.addItem(img)

            # interactive histogram
            hist = pg.HistogramLUTItem()
            # link the histogram to the image
            hist.setImageItem(img)
            # set histogram gradient
            hist.gradient.restoreState(
                {'mode': 'rgb',
                 'ticks': [(0.5, (0, 182, 188, 255)),
                           (1.0, (246, 111, 0, 255)),
                           (0.0, (75, 0, 113, 255))]})

            win.addItem(hist)


            plots.append(plt)
            plot_imgs.append(img)

            win.nextRow()

        ch_inds = list(channels)


    elif isinstance(channels, dict):
        for ch_ind, ch_name in channels.items():

            plt = win.addPlot(title=f'{ch_name} channel')
            plt.setMouseEnabled(x=False, y=False)

            img = pg.ImageItem()
            plt.addItem(img)

            # interactive histogram
            hist = pg.HistogramLUTItem()
            # link the histogram to the image
            hist.setImageItem(img)
            # set histogram gradient
            hist.gradient.restoreState(
                {'mode': 'rgb',
                 'ticks': [(0.5, (0, 182, 188, 255)),
                           (1.0, (246, 111, 0, 255)),
                           (0.0, (75, 0, 113, 255))]})

            win.addItem(hist)


            plots.append(plt)
            plot_imgs.append(img)

            win.nextRow()

        ch_inds = list(channels.keys())

    else:
        raise ValueError('`channels` argument should be list, tuple, or dict')


    # label bottom plot's x-axis
    plots[-1].setLabel('bottom', 'time', units='s')


    # process initialization events
    app.processEvents()

    # open stream
    inlet.open_stream()

    running = True
    while running:
        chunk, _ = inlet.pull_chunk(max_samples=buffer_length)
        if np.size(chunk) != 0:  # Check if new data available
            buffer = np.append(buffer, np.array(chunk), axis=0)

            # clip to buffer_length
            buffer = buffer[-buffer_length:]

            for ch_ind, plot_img in zip(ch_inds, plot_imgs):
                freqs, _, Sxx = signal.spectrogram(
                    buffer[:, ch_ind], s_rate, mode='magnitude')

                # FIXME: temp fix of outliers
                Sxx_clipped = np.clip(Sxx, 0, 100)
                plot_img.setImage(Sxx_clipped)

                #plot_curve.setData(y=buffer[:, ch_ind], x=ts)

            app.processEvents()

        if not win.isVisible():
            running = False
            app.quit()



import setuptools


with open("readme.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="neurol",
    version="0.0.3",
    author="Awni",
    author_email="awni.altabaa@queensu.ca",
    description="A package for modularly implenting Brain-Computer Interfaces",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="BCI Brain-Computer Interface Neurotechnology Neuroscience EEG",
    install_requires=[
            "numpy",
            "matplotlib",
            "biosppy",
            "ble2lsl",
            "pylsl",
            "scipy",
            "scikit_learn",
            "Keras"
    ],
    include_package_data=True,
    url='https://github.com/Awni00/neurol',
    project_urls={'Documentation': 'https://neurol.readthedocs.io/',
        'Source':'https://github.com/Awni00/neurol',
        'Tracker':'https://github.com/Awni00/neurol/issues'},
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



'''
Module containing functions for performing classification, via machine learning
models or otherwise, related to Brain-Computer Interface applications.
'''

import numpy as np



def get_channels(signal, channels, device='muse'):
    '''
    Returns a signal with only the desired channels.

    Arguments:
        signal: a signal of shape [n_samples, n_channels]
        channels: an array of the str names of the desired channels.
            returned in given order.
        device: str name of the device.

    Returns:
        numpy array of signal with shape [n_channels, n_desired_channels].
        Includes only the selected channels in the order given.
    '''

    # check device; each device has its own ch_ind dictionary corresponding to
    # its available channels
    if device == 'muse':
        ch_ind_muse = {'TP9': 0, 'AF7': 1, 'AF8': 2, 'TP10': 3}
        return_signal = np.array([np.array(signal)[:, ch_ind_muse[ch]]
                                  for ch in channels]).T

    return return_signal


def softmax_predict(input_, predictor, thresh=0.5):
    '''
    Consolidates a softmax prediction to a one-hot encoded prediction.

    Arguments:
        input_: the input taken by the predictor
        predictor: function which returns a softmax prediction given an input_
        thresh: the threshold for a positive prediction for a particular class.

    '''

    pred = np.array(predictor(input_))

    return (pred >= thresh).astype(int)


def encode_ohe_prediction(prediction):
    '''
    Returns the index number of the positive class in a
    one-hot encoded prediction.
    '''
    return np.where(np.array(prediction) == 1)[0][0]


def decode_prediction(prediction, decode_dict):
    '''
    Returns a more intelligible reading of the prediction
    based on the given decode_dict
    '''
    return decode_dict[prediction]


def threshold_clf(features, threshold, clf_consolidator='any'):
    '''
      Classifies given features based on a given threshold.

      Arguments:
          features: an array of numerical features to classify
          threshold: threshold for classification. A single number, or an
            array corresponding to `features` for element-wise comparison.
          clf_consolidator: method of consolidating element-wise comparisons
            with threshold into a single classification.
              'any': positive class if any features passes the threshold
              'all': positive class only if all features pass threshold
              'sum': a count of the number of features which pass the threshold
              function: a custom function which takes in an array of booleans
                and returns a consolidated classification

      Returns:
          classification for the given features. Return type `clf_consolidator`.
    '''

    try:
        label = np.array(features) > np.array(threshold)
    except ValueError as v_err:
        print("Couldn't perform comparison between features and thresholds."
              "Try a different format for the threshold.")
        raise v_err

    # consolidate binary label array into single classification
    if clf_consolidator == 'any':
        label = np.any(label)

    elif clf_consolidator == 'all':
        label = np.all(label)

    elif clf_consolidator == 'sum':
        label = np.sum(label)

    elif callable(clf_consolidator):
        try:
            label = clf_consolidator(label)
        except TypeError as t_err:
            print("Couldn't consolidate classification with `clf_consolidator`")
            raise t_err

    return label
